package com.processmininguc.experiments;

import java.io.File;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.List;

import org.deckfour.xes.in.XesXmlParser;
import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.util.Pair;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.ptml.importing.PtmlImportTree;

import au.com.bytecode.opencsv.CSVWriter;

@Plugin(name = "Configurable Processes Experiments", parameterLabels = { }, 
	    returnLabels = { "Useless Output" }, returnTypes = { String.class }, help = "Systematic Experiments on Configurable Processes")
public class ConfigurableProcessesExperimentsPlugin{


	//TODO Remove @UITopiaVariant when publish
	@UITopiaVariant(affiliation = "Universidad Catolica de Chile", 
				author = "Mauricio Arriagada-Benitez", 
				email = "marriagb@uc.cl")
	@PluginVariant(variantLabel = "Configurable Processes Experiments", requiredParameterLabels = {})
	public String run(PluginContext context) throws Exception {
		
		//============================================================
		//CONFIGURATION
		String baseFolder = "/Users/jmunoz/Dropbox/benchmarks/configurableprocesses/test1/";
		String datasetsFolder = baseFolder+"datasets/";
		String resultsFolder = baseFolder+"results/";
		long experimentNumber = System.currentTimeMillis();
		String resultsPath = resultsFolder+experimentNumber+".csv";
		int numRuns = 1;
		//============================================================

		//============================================================
		//FIELDS
		String RUN = "run";
		String LOG = "log";
		String TREE = "tree";
		String ALGORITHM = "algorithm";
		String PAR_FIT = "parfit";
		String PAR_PRE = "parpre";
		String PAR_GEN = "pargen";
		String PAR_SIM = "parsim";
		String INI_TIME = "initime";
		String END_TIME = "endtime";
		String TOTAL_TIME = "totaltime";
		String FIT = "fit";
		String PRE = "pre";
		String GEN = "gen";
		String SIM = "sim";
		//============================================================
		
		
		
		String datasetsPath = "/Users/jmunoz/Dropbox/benchmarks/configurableprocesses/test1/datasets";
		List<Pair<File,File>> files = getTreesAndLogsFiles(datasetsPath);
		
		CSVWriter writer = new CSVWriter(new FileWriter(new File(resultsPath)));

		return "Experiments Finished";  
	}
	
	public List<Pair<File,File>> getTreesAndLogsFiles(String datasetsPath){
		List<Pair<File,File>> list = new ArrayList<Pair<File,File>>();
		File baseFile = new File(datasetsPath);
		for (File modelFile: baseFile.listFiles()){
			if (!modelFile.isDirectory() && modelFile.getName().endsWith(".ptml")){
				String processName = modelFile.getName().replace(".ptml", "");
				for (File logFile: baseFile.listFiles()){
					if (!logFile.isDirectory() &&
						logFile.getName().endsWith(".xes") &&
						logFile.getName().startsWith(processName)){
						list.add(new Pair<File,File>(modelFile,logFile));
					}
				}
			}
		}
		return list;	
	}
	
	public ProcessTree getProcessTree(PluginContext context, File file) throws Exception{
		PtmlImportTree ptmlImport = new PtmlImportTree();
		ProcessTree tree = (ProcessTree) ptmlImport.importFile(context, file);
		return tree;
	}
	
	public XLog getLog(PluginContext context, File file) throws Exception{
		XesXmlParser parser = new XesXmlParser();
		XLog log = parser.parse(file).get(0);
		return log;
	}


}
