/**
 * 
 */
package org.processmininguc.configurableprocesses.algorithms;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;
import java.util.UUID;

import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.info.impl.XLogInfoImpl;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.plugins.etm.CentralRegistry;
import org.processmining.plugins.etm.model.narytree.Configuration;
import org.processmining.plugins.etm.model.narytree.NAryTree;
import org.processmining.plugins.etm.model.narytree.conversion.NAryTreeToProcessTree;
import org.processmining.plugins.etm.model.narytree.conversion.ProcessTreeToNAryTree;
import org.processmining.processtree.Block;
import org.processmining.processtree.Edge;
import org.processmining.processtree.Node;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.configuration.controlflow.impl.ControlFlowConfigurationImpl;
import org.processmining.processtree.configuration.controlflow.impl.ControlFlowConfigurationInstantiaterImpl;
import org.processmining.processtree.impl.ProcessTreeImpl;
import org.processmininguc.configurableprocesses.parameters.GreedyParameters;

/**
 * @author mauro
 *
 */
public class GreedyAlgorithm {


	/**
	 * The method that implements your algorithm.
	 * 
	 * @param input1 The first input.
	 * @param input2 The second input.
	 * @param parameters The parameters to use.
	 * @return The output.
	 */
	public ProcessTree apply(PluginContext context, ProcessTree tree, XLog log, GreedyParameters parameters,
			final ConfigurableProcessesResult confProcResults) {
		/**
		 * Put your algorithm here, which computes an output form the inputs provided the parameters.
		 */
		long startTime = System.currentTimeMillis();
		
		//ProcessTree bestProcessTree = TreeTests.createRandomProcessTree();
		ProcessTree bestProcessTree = greedyAlgorithm(context, tree, log, parameters, confProcResults);
		//ProcessTree treeTest = greedyTests(tree);  //Configuration test
		long endTime   = System.currentTimeMillis();
		String finalTime = showElapsedTime(startTime,endTime);

		confProcResults.setTimeElapsed(finalTime);
		return bestProcessTree;
	}

	public static ProcessTree greedyAlgorithm(PluginContext context, ProcessTree tree, XLog log, GreedyParameters parameters, final ConfigurableProcessesResult confProcResults){
		//General variable purposes
		Collection<String> eventList = null;
		XLog reducedLog = null;
		ProcessTree subTree = null;
		double bestFitness=0.0, bestOverallFitness=0.0, fitness=0.0;
		ProcessTree convertedTree = null, bestTree=null;
		Configuration newConfiguration=null;

		Map<String,String> finalConfiguration = new HashMap<String,String>(); //this holds the final configuration
		
		List<UUID> uuidsNodeList = null;	//simple list of UUID from nodes
		boolean go = true;		
				
		//List of all general configurations 
		Map<UUID,List<Byte>> allUUIDsNodeConfsMap = new HashMap<UUID,List<Byte>>();
		allUUIDsNodeConfsMap = getAllConfigurations(tree, allUUIDsNodeConfsMap);
		
		Map<UUID,List<Byte>> uuidsNodeConfsMap = new HashMap<UUID,List<Byte>>();
		uuidsNodeConfsMap = getAvailableConfigurations(tree, allUUIDsNodeConfsMap);
				
		confProcResults.setConfigurationSize(allUUIDsNodeConfsMap.size());
		System.out.println("\n##### Start Greedy ##### " + allUUIDsNodeConfsMap.size() + " - confs");
		//iterate over all configuration points. For each configuration point we check all configuration
		while(go){
			if (tree.size() > 1 && !allUUIDsNodeConfsMap.isEmpty()){
				Map<UUID,List<Byte>> uuidNodeConfMap = new HashMap<UUID,List<Byte>>();
				uuidNodeConfMap = getProperConfiguration(tree, allUUIDsNodeConfsMap);
				//this loop iterates only 1 time
				for (Map.Entry<UUID, List<Byte>> entry : uuidNodeConfMap.entrySet()) {
					System.out.println("**** key : " + entry.getKey() + " | values : "+entry.getValue());
					//add configuration activity name to a HashMap with no value
					String actName = tree.getNode(entry.getKey()).getName();
					//finalConfiguration.put(actName,null);
					System.out.println("act name : "+actName);
					
					eventList = new ArrayList<String>();
					Node node = tree.getNode(entry.getKey());
					Block block = node.getIncomingEdges().get(0).getSource();
					finalConfiguration.put(node.toString(),null);
					//System.out.println("type : "+ node.toStringShort());
					

					//get a list of events from a subtree
					getListEventsFromSubtree(block, eventList);
					reducedLog = filterLog(log, XLogInfoImpl.NAME_CLASSIFIER, eventList) ;

					//Special case when having a loop operator
					//ver también poner a TAU las actividades que son configurables 
					
					//get a subtree assigning a new root to the tree
					List<Node> blockList = block.getChildren();
					subTree = new ProcessTreeImpl(tree);
					if (node.isLeaf()){
						subTree.setRoot(block);						
						System.out.println("children : " + block.getChildren().size() + " / " + subTree.toString());
					}else{
						subTree.setRoot(node);
						System.out.println("children : " + block.getChildren().size() + " / " + subTree.toString());
					}

					for (int i=0 ; i < blockList.size() ; i++) {
						if ( Arrays.asList(blockList).contains(blockList.get(i)) )
							System.out.println("child : " + blockList.get(i));
							//set to tau all configurable node children
					}

					CentralRegistry registry = new CentralRegistry(reducedLog, XLogInfoImpl.NAME_CLASSIFIER, new Random());
					ProcessTreeToNAryTree converter = new ProcessTreeToNAryTree(registry.getEventClasses());
					NAryTree simpleNAryTree = converter.convert(subTree);

					uuidsNodeList = new ArrayList<UUID>();
					uuidsNodeList = getTreeListPreOrder(subTree.getRoot(), uuidsNodeList);	//pre-order list of tree nodes 

					Byte bestConf = null;

					//iterate over all configurations from a configuration point to get the best configuration
					bestConf = getBestConfigurationFromConfigurationPoint(entry, simpleNAryTree, uuidsNodeList, registry, context, reducedLog,parameters,confProcResults);
					System.out.println("best conf : " + bestConf);

					//Apply the best configuration to our main tree
					tree = applyBestConfigurationToOriginalTree(tree, entry,bestConf);

					//best local configuration
					System.out.println("\n==== Show Best Local Configuration ====");
					System.out.println(" best Tree : " + tree + "\n best Configuration : " + bestConf);

					//updating value of a configuration activity name of a HashMap
					if (bestConf.equals(Configuration.BLOCKED)){
						finalConfiguration.put(actName,"B");						
					}
					else if(bestConf.equals(Configuration.HIDDEN)){
						finalConfiguration.put(node.toString(),"H");						
					}
					else if(bestConf.equals(Configuration.NOTCONFIGURED)){ 
						finalConfiguration.put(node.toString(),"E");
					}
				}
			}else{
				go = false;
			}
		}

		//best Global configuration
		System.out.println("\n==== Show Best Global Configuration ====");
        System.out.println(" best Tree : " + tree);
		Double conformance = ConformanceMetric.getNAryTreeConformance(context,tree,log,parameters,XLogInfoImpl.NAME_CLASSIFIER,confProcResults);
       	System.out.println("Conformance : "+ conformance);
		System.out.println("\n##### End Greedy #####");
		
		//printing a map with act name and configuration option
		System.out.println("list of final configuration : " + finalConfiguration);
		confProcResults.setFinalConf(finalConfiguration);
		
		confProcResults.setConformance(conformance);
//		confProcResults.setBestConfiguration(bestConf);
        return tree;
	}
	private static Map<UUID, List<Byte>> getAvailableConfigurations(ProcessTree tree, Map<UUID, List<Byte>> allUUIDsNodeConfsMap) {
		// TODO Auto-generated method stub
		Collection<Node> treeNodes = tree.getNodes();
		Map<UUID, List<Byte>> temp = new HashMap<UUID, List<Byte>>();
		
		for(Node node : treeNodes){
			if(allUUIDsNodeConfsMap.containsKey(node.getID())){
				temp.put(node.getID(), allUUIDsNodeConfsMap.get(node.getID()));
			}
		}
		System.out.println("show all confs : "+temp);
		return temp;
	}

	//This applies configurations to the tree and then checks if there exits available configuration points to return
	private static Map getProperConfiguration(ProcessTree tree, Map<UUID, List<Byte>> uuidsNodeConfsMap) {
		// TODO Auto-generated method stub
		Map<UUID, List<Byte>> mapTemp = new HashMap<UUID, List<Byte>>();
		int properConfSize = 999; //used to get the smaller subtree under a config node as a root
		UUID bestConfId = null;
		List<Byte> bestValues = null;
		for (Map.Entry<UUID, List<Byte>> entry : uuidsNodeConfsMap.entrySet()) {
			Node node = tree.getNode(entry.getKey());
			Block block = node.getIncomingEdges().get(0).getSource();
			
			ProcessTree subTree = new ProcessTreeImpl(tree);
			if (node.isLeaf()){
				subTree.setRoot(block);
			}else{
				subTree.setRoot(node);
			}
			
			if(subTree.size() < properConfSize){
				bestConfId = entry.getKey();
				bestValues = entry.getValue();
				properConfSize = subTree.size();
			}
		}
		mapTemp.put(bestConfId,bestValues);
		uuidsNodeConfsMap.remove(bestConfId); //eliminates a configuration point after have selected it
		return mapTemp;
	}

	//TO-DO, This function needs to be fixed since we are going to apply configuration from ProcessTree package and not from ETM
	private static ProcessTree applyBestConfigurationToOriginalTree(ProcessTree tree, Entry<UUID, List<Byte>> entry, Byte bestConf) {
		ControlFlowConfigurationImpl cfc = new ControlFlowConfigurationImpl();
		ControlFlowConfigurationInstantiaterImpl cfci = new ControlFlowConfigurationInstantiaterImpl();
		ProcessTree ptInstantiated = null;
		UUID id = tree.getEdge(tree.getNode(entry.getKey()).getIncomingEdges().get(0).getID()).getID();

		if (bestConf.equals(Configuration.BLOCKED)){
			cfc.addBlockedEdge(id);
			ptInstantiated =  cfci.instantiate(tree, cfc);
		}
		else if(bestConf.equals(Configuration.HIDDEN)){
			cfc.addHiddenEdge(id);
			ptInstantiated = cfci.instantiate(tree, cfc);
		}
		else if(bestConf.equals(Configuration.NOTCONFIGURED)){
			cfc.addNonBlockedEdge(id);
			cfc.addNonHiddenEdge(id);
			//cfc.addNonChangeableNode(id);
			ptInstantiated = cfci.instantiate(tree, cfc);
		}
		return ptInstantiated;
	}

	private static Byte getBestConfigurationFromConfigurationPoint(Entry<UUID, List<Byte>> entry, NAryTree simpleNAryTree,
									List<UUID> uuidsNodeList, CentralRegistry registry, PluginContext context, XLog reducedLog, GreedyParameters parameters, final ConfigurableProcessesResult confProcResults) {
		// TODO Auto-generated method stub
		Configuration c;
		Byte bestConf = null;
		Double bestConformance=0.0, conformance;
		//traverse the configuration of a configuration point
		for(Byte entryConf : entry.getValue()){
			System.out.println("--- configuration " + entryConf + " ---");
			c = new Configuration(simpleNAryTree.size());
			c = setConfiguration(uuidsNodeList,entry.getKey(),entryConf, c);
			
			simpleNAryTree.addConfiguration(c);
			NAryTree res = simpleNAryTree.applyConfiguration(0);
			
			ProcessTree convertedTree =  NAryTreeToProcessTree.convert(res, registry.getEventClasses());
			
			conformance = ConformanceMetric.getNAryTreeConformance(context,convertedTree,reducedLog,parameters, XLogInfoImpl.NAME_CLASSIFIER,confProcResults);
			simpleNAryTree.removeConfiguration(0);
        	
        	//first we compare the fitness metric. In case the fitness are equal, we compare the overall fitness
        	if (conformance > bestConformance){
        		bestConformance = conformance;
				bestConf = entryConf;
        	}
        	System.out.println("Conformance : "+ conformance);
		}
		System.out.println("Best conformance : " + bestConformance);
		return bestConf;
	}

	private static Map<UUID, List<Byte>> getAllConfigurations(ProcessTree tree, Map<UUID, List<Byte>> uuidsNodeConfsMap) {
		// TODO Auto-generated method stub
		for (Edge edge :tree.getEdges()){
			List<Byte> bytelist = new ArrayList<Byte>();
			if (edge.isBlockable())
				bytelist.add(Configuration.BLOCKED);
			if(edge.isHideable())
				bytelist.add(Configuration.HIDDEN);
			if(edge.isBlockable() || edge.isHideable()){				
				bytelist.add(Configuration.NOTCONFIGURED);
				uuidsNodeConfsMap.put(tree.getEdge(edge.getID()).getTarget().getID(),bytelist);
			}
		}
		//Map<UUID,List<Byte>> finaluuidsNodeConfsMap = new HashMap<UUID,List<Byte>>();
		//for (Map.Entry<UUID, List<Byte>> entry : uuidsNodeConfsMap.entrySet()){
		//	if (noParent(tree,entry.getKey())){
		//		finaluuidsNodeConfsMap.add(tree.getEdge(entry.getKey()).getTarget().getID(),null);
		//	}		
		//}
		return uuidsNodeConfsMap;
	}

	private static boolean noParent(ProcessTree tree, UUID key) {
		// TODO Auto-generated method stub
		if (tree.getNode(key).isRoot()){
			return true;
		}
		List<Edge> n = tree.getNode(key).getIncomingEdges();
		return noParent(tree, n.get(0).getID());
	}

	private static Configuration setConfiguration(List<UUID> uuidsNodeList, UUID key, Byte entryConf, Configuration c) {
		// TODO Auto-generated method stub
		if (entryConf.equals(Configuration.BLOCKED))
			c.conf[uuidsNodeList.indexOf(key)]=Configuration.BLOCKED;
		else if(entryConf.equals(Configuration.HIDDEN))
			c.conf[uuidsNodeList.indexOf(key)]=Configuration.HIDDEN;
		else if(entryConf.equals(Configuration.NOTCONFIGURED))
			c.conf[uuidsNodeList.indexOf(key)]=Configuration.NOTCONFIGURED;

		return c;
	}

	private static void getListEventsFromSubtree(Node node, Collection<String> listEvents) {
		// TODO Auto-generated method stub
		if (node.isLeaf())
			listEvents.add(node.getName());
		else{
			Block b = (Block) node;
			for (Node nod: b.getChildren())
				getListEventsFromSubtree(nod, listEvents);
		}
	}

	public static List<UUID> getTreeListPreOrder(Node node, List<UUID> uuidList){
		if (node.isLeaf()){
			uuidList.add(node.getID());
			return uuidList;
		}else{
			uuidList.add(node.getID());
			Block b = (Block) node;
			for (Node nod: b.getChildren()){
				getTreeListPreOrder(nod, uuidList);
			}
		}
		return uuidList;
	}
	
	public static XLog filterLog(XLog eventLog, XEventClassifier classifier, Collection<String> eventList) {
		XLog filteredLog = (XLog) eventLog.clone();
		Iterator<XTrace> itTrace = filteredLog.iterator();
		while (itTrace.hasNext()) {
			XTrace trace = itTrace.next();
			Iterator<XEvent> itEvent = trace.iterator();
			while (itEvent.hasNext()) {
				XEvent event = itEvent.next();
				String id = classifier.getClassIdentity(event);
				if (!eventList.contains(id)) {
					itEvent.remove();
				}
			}
		}
		return filteredLog;
	}
	
	
	private String showElapsedTime(long startTime, long endTime) {
		// TODO Auto-generated method stub
		long time = endTime - startTime;
		long milli  = time % 1000;
		long second = (time / 1000) % 60;
		long minute = (time / (1000 * 60)) % 60;
		long hour = (time / (1000 * 60 * 60)) % 24;
		String timeFormatted = String.format("%02d:%02d:%02d:%d", hour, minute, second, milli);
		System.out.println("Time elapsed : "+timeFormatted);
		return timeFormatted;
	}

	//public static void checkParents(Map<UUID, List<Byte>> allUUIDsNodeConfsMap){
	//	System.out.print("initial checking...\n");
	//	for (Map.Entry<UUID, List<Byte>> entry : allUUIDsNodeConfsMap.entrySet()) {
	//		System.out.println(entry.getKey());
	//		System.out.println(entry);
	//	}	
	//	System.out.print("final checking...");
	//}
	
	
/*	public static ProcessTree greedyTests(ProcessTree tree){
		ControlFlowConfigurationImpl cfc = new ControlFlowConfigurationImpl();
		UUID id =  UUID.fromString("18da2637-9c4e-44d5-93ea-5461fbcfcbef");
		cfc.addHiddenEdge(id);
		ControlFlowConfigurationInstantiaterImpl cfg = new ControlFlowConfigurationInstantiaterImpl();
		ProcessTree ptInstantiate = cfg.instantiate(tree, cfc);
		System.out.println("ptInstantiate : " + ptInstantiate);
		
		return ptInstantiate;
	}
*/	
	
}
