package org.processmininguc.configurableprocesses.parameters;

import org.deckfour.xes.model.XLog;
import org.processmining.basicutils.parameters.impl.PluginParametersImpl;
import org.processmining.processtree.ProcessTree;

public class GeneticParameters extends PluginParametersImpl {

	private boolean verbose;
	private String resultName;
	public int maxAllowedEvolutions;
	public int populationSize;
	public int fitnessWeigth, precisionWeigth, generalizationWeigth, simplicityWeigth;
	
	public GeneticParameters(ProcessTree tree, XLog log) {
		super();
		setVerbose(false);
		setTryConnections(false);
		setResultName("Configuration for "+tree.getName());
		//DEFAULT VALES
		maxAllowedEvolutions = 20;
		populationSize = 10;
		fitnessWeigth = 90;
		precisionWeigth = 10;
		generalizationWeigth = 0;
		simplicityWeigth = 0;
	}

	public GeneticParameters(GeneticParameters parameters) {
		super(parameters);
		setVerbose(parameters.isVerbose());
		setResultName(parameters.getResultName());
		setTryConnections(parameters.isTryConnections());
	}
	
	public boolean equals(Object object) {
		if (object instanceof GeneticParameters) {
			GeneticParameters parameters = (GeneticParameters) object;
			return super.equals(parameters) &&
					isVerbose() == parameters.isVerbose() &&
					getResultName().equals(parameters.getResultName());
		}
		return false;
	}
	
	public boolean isVerbose() {
		return verbose;
	}

	public void setVerbose(boolean verbose) {
		this.verbose = verbose;
	}

	public String getResultName() {
		return resultName;
	}

	public void setResultName(String resultName) {
		this.resultName = resultName;
	}

	public String toString() {
		return "(" + getResultName() + "," + isVerbose() + ")";
	}
	
	//new initial Genetic Algorithm parameters
	public int getMaxAllowedEvolutions() {
		return maxAllowedEvolutions;
	}

	public void setMaxAllowedEvolutions(int maxAllowedEvolutions) {
		this.maxAllowedEvolutions = maxAllowedEvolutions;
	}

	public int getPopulationSize() {
		return populationSize;
	}

	public void setPopulationSize(int populationSize) {
		this.populationSize = populationSize;
	}
	
	public void setFitnessWeigth(int fitness){
		this.fitnessWeigth = fitness;
	}
	
	public void setPrecisionWeigth(int precision){
		this.precisionWeigth = precision;
	}
	
	public void setGeneralizationWeigth(int generalization){
		this.generalizationWeigth = generalization;
	}
	
	public void setSimplicityWeigth(int simplicity){
		this.simplicityWeigth = simplicity;
	}
	
	public int getFitnessWeigth(){
		return fitnessWeigth;
	}
	
	public int getPrecisionWeigth(){
		return precisionWeigth;
	}
	
	public int getGeneralizationWeigth(){
		return generalizationWeigth;
	}
	
	public int getSimplicityWeigth(){
		return simplicityWeigth;
	}
}
